#include <direct.h>    // DirectDraw definitions 
#include <d3drmwin.h>  // D3DRM definitions 
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <dsound.h>
#include <crtdbg.h>

#include "RubiksSoundSet.h"
#include "wave.h"
#include "debug.h"


// Constructor
SoundSet::SoundSet( )
{
	SoundSet( NULL, NULL );
}

// Constructor
SoundSet::SoundSet(HWND hwnd, GUID *pguid)
{
	if( InitDSound( hwnd, pguid ) != TRUE )
		DISPLAYMSG("Direct Sound not initialized" );

	BOOL success = Init3DListener();
	_ASSERT(success);
	
	m_vPos.x = 0.0;
	m_vPos.y = 0.0;
	m_vPos.z = 0.0;

	m_vOrientation.x = 0.0;
	m_vOrientation.y = 0.0;
	m_vOrientation.z = 0.0;
}

//Destructor
SoundSet::~SoundSet()
{
	PBUFFERINFO pTempBufferInfo;	// temporary bufferInfo struct;
	
	// Iterate through buffer list and release all sound buffers
	while( BufferList.size() > 0 )
	{
		pTempBufferInfo = BufferList.front();	// Get first info struct
		ReleaseBufferInfo( pTempBufferInfo );	// Release the bufferinfo and buffers and remove elt from list
	}

	// Release all other sound components
	RELEASE( lpds );
	lpds = NULL;
	RELEASE( lpdsb );
	lpdsb = NULL;
	RELEASE( lpdsbPrimary );
	RELEASE( lp3dListener );
}

/*****************************************************************************************	
 *	InitDSound
 *
 *	The first step is to create the DirectSound object, establish a cooperative level, and 
 *	set the primary buffer format. All this is done in the InitDSound function.
 *
 *	The function takes two parameters: the main window handle and a pointer to the GUID of 
 *	the sound device. In most cases you will pass NULL as the second parameter, indicating 
 *	the default device, but you may obtain a GUID by device enumeration.
 ******************************************************************************************/
BOOL SoundSet::InitDSound(HWND hwnd, GUID *pguid)
{
   // Create DirectSound
    if FAILED(DirectSoundCreate(pguid, &lpds, NULL)) 
        return FALSE;
 
    // Set co-op level
    if FAILED(IDirectSound_SetCooperativeLevel(
            lpds, hwnd, DSSCL_PRIORITY))
        return FALSE;
 
/*	You have set the priority cooperative level in order to be able to set the format of the 
	primary buffer. If you don't change the default format, output will be in the 8-bit, 
	22 kHz format regardless of the format of the input. Setting the primary buffer to a 
	higher format can do no harm, because even if the secondary buffers are in a lower 
	format, the samples will be converted automatically by DirectSound. Note also that 
	there's no danger of the call to IDirectSoundBuffer::SetFormat failing because the 
	hardware doesn't support the higher format. DirectSound will simply set the closest
	available format.

	To set the format of the primary buffer, you first describe it in the global 
	DSBUFFERDESC structure, then pass this description to the IDirectSound::CreateSoundBuffer 
	method.
*/
    // Obtain primary buffer
    ZeroMemory(&dsbdesc, sizeof(DSBUFFERDESC));
    dsbdesc.dwSize = sizeof(DSBUFFERDESC);
    dsbdesc.dwFlags = DSBCAPS_PRIMARYBUFFER | DSBCAPS_CTRL3D;
    if FAILED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsbPrimary, NULL))
    {
		DISPLAYMSG("Failed to create primary buffer");
		return FALSE;
	}

 
/*	Once you have a primary buffer object, you describe the desired wave format and pass 
	the description to the IDirectSoundBuffer::SetFormat method:
*/

    // Set primary buffer format
    WAVEFORMATEX wfx;
    memset(&wfx, 0, sizeof(WAVEFORMATEX)); 
    wfx.wFormatTag = WAVE_FORMAT_PCM; 
    wfx.nChannels = 2; 
    wfx.nSamplesPerSec = 44100; 
    wfx.wBitsPerSample = 16; 
    wfx.nBlockAlign = wfx.wBitsPerSample / 8 * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
 
    if( FAILED( lpdsbPrimary->SetFormat(&wfx) ) )
	{
		DISPLAYMSG("Failed setting format on primary buffer");
		return FALSE;
	}

    return TRUE;
}


/*******************************************************************************************
 *
 *	SetUpStaticBuffer( char* pcSoundFileName )
 *
 *	This function takes care of creating and locking a static buffer just large enough to 
 *  contain entire wav file and then reads and writes the wave file to that buffer.  All
 *  that remains to be done is to play the buffer.
 *
 ******************************************************************************************/
BOOL SoundSet::SetUpBuffer(char *pcSoundFileName, BUFFER_TYPE eBufType )
{
	if (WaveOpenFile( pcSoundFileName, &hmmio, &pwfx, &mmckinfoParent) != 0)
	{
		DISPLAYMSG("Failed to open wave file");
		return FALSE;
	}
 
/*	Note that the wrapper functions all return zero if successful.

	The next step is to call the WaveStartDataRead function, causing the file pointer 
	to descend to the data chunk. This function also fills in the MMCKINFO structure for 
	the data chunk, so that you know how much data is available:
*/
	if (WaveStartDataRead(&hmmio, &mmckinfoData, &mmckinfoParent) != 0)
	{
		DISPLAYMSG("Failed in WaveStartDataRead on primary buffer");
		return FALSE;
	}
 
/*	The application can now begin copying data from the file to a secondary sound buffer. 
	Normally you don't create the sound buffer until you have obtained the size of the 
	data chunk and the format of the wave. The following code creates a static buffer just
	large enough to hold all the data in the file. 
*/

/*	It is assumed that lpds is a valid pointer
	to the DirectSound object. 
*/
	if( CreateBuffer( eBufType ) != TRUE ) 
	{
		DISPLAYMSG("Failed to create correct buffer");
		return FALSE;
	}
 
/*	Because in this case the application is not streaming the data but simply filling a
	static buffer, the entire buffer is locked from the beginning. There is no wraparound, 
	so only a single pointer and byte count are required.
*/
	LPVOID lpvAudio1;
	DWORD  dwBytes1;
 
	if FAILED(lpdsb->Lock(
			0,              // Offset of lock start
			0,              // Size of lock; ignored in this case
			&lpvAudio1,     // Address of lock start
			&dwBytes1,      // Number of bytes locked
			NULL,           // Wraparound start; not used
			NULL,			// Wraparound size; not used
			DSBLOCK_ENTIREBUFFER))  // Flag
	{
		// Error handling
		WaveCloseReadFile(&hmmio, &pwfx);
		DISPLAYMSG("Failed to lock secondary static buffer");
		return FALSE; 
	}
 
/*	The WaveReadFile function in Wave.c copies the data from the file to the buffer pointer 
	and returns zero if successful. 
*/
	UINT cbBytesRead;
 
	if (WaveReadFile(
			hmmio,              // file handle
			dwBytes1,           // no. of bytes to read
			(BYTE *) lpvAudio1, // destination
			&mmckinfoData,      // file chunk info 
			&cbBytesRead))      // actual no. of bytes read
	{
		// Handle failure on non-zero return
		WaveCloseReadFile(&hmmio, &pwfx);
		DISPLAYMSG("Failed to read wave file into secondary static buffer");
		return FALSE; 
	}
 
/*	Finally, the application unlocks the buffer and closes the wave file:
*/
	lpdsb->Unlock(lpvAudio1, dwBytes1, NULL, 0);
	WaveCloseReadFile(&hmmio, &pwfx);
 
/*	For a streaming buffer, you would typically call WaveReadFile at regular intervals 
	determined by the current play position. (See Play Buffer Notification.) If the locked 
	portion of the buffer wrapped around, of course, you would call WaveReadFile once for 
	each segment of the lock.
*/
	return TRUE;
}




BOOL SoundSet::Play( SOUND eSound, D3DVECTOR vPos, D3DVECTOR vOrient )
{
	// Record position and orientation of where sound will be played from
	m_vPos = vPos;
	m_vOrientation = vOrient;

	// Figure out how to set up buffer based on particular sound
	switch( eSound )
	{
	case FIRE_WORKS:
		if( SetUpBuffer( "fireworks.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case SKID:	
		if( SetUpBuffer( "skid.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case BOUNCE:	
		if( SetUpBuffer( "bounce.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case SHIELD:	
		if( SetUpBuffer( "shield.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case CANNON:	
		if( SetUpBuffer( "cannon.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case PROJECTILE:
		if( SetUpBuffer( "shield.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, DSBPLAY_LOOPING );
		break;

	case HERO_SKID:	
		if( SetUpBuffer( "skid.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_BOUNCE:	
		if( SetUpBuffer( "bounce.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_SHIELD:	
		if( SetUpBuffer( "shield.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_CANNON:	
		if( SetUpBuffer( "cannon.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case WIN:	
		if( SetUpBuffer( "win.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case LOSE:	
		if( SetUpBuffer( "lose.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case SBOUNCE:	
		if( SetUpBuffer( "sbounce.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_WIN:	
		if( SetUpBuffer( "win.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_LOSE:	
		if( SetUpBuffer( "lose.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;

	case HERO_SBOUNCE:	
		if( SetUpBuffer( "sbounce.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case SCREAM:	
		if( SetUpBuffer( "ahhhhh.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case GREATBABY:
		if( SetUpBuffer( "great.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case BANG:
		if( SetUpBuffer( "bang.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case HERO_SCREAM:	
		if( SetUpBuffer( "ahhhhh.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case HERO_GREATBABY:
		if( SetUpBuffer( "great.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case HERO_BANG:
		if( SetUpBuffer( "bang.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case HERO_CLANK:
		if( SetUpBuffer( "metalpipe.wav", STATIC ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;
	case CLANK:
		if( SetUpBuffer( "metalpipe.wav", STATIC_3D ) != TRUE ) goto failed_buffer;	
		lpdsb->Play( 0, 0, 0 ); // The arguments are flags that all happen to be zero (not coordinates)
		break;


	default:
		goto failed_buffer;
	}

	
	return TRUE;

failed_buffer:
	DISPLAYMSG("Failed to set up buffer correctly");
	return FALSE;
}

BOOL SoundSet::Init3DListener()
{
/* To obtain a pointer to an IDirectSound3DListener interface, you must first create a 
   primary 3-D sound buffer. Do this by using the IDirectSound::CreateSoundBuffer method, 
   specifying the DSBCAPS_CTRL3D and DSBCAPS_PRIMARYBUFFER flags in the dwFlags member of 
   the accompanying DSBUFFERDESC structure. Then, use the IDirectSoundBuffer::QueryInterface
   method on the resulting buffer to obtain a pointer to an IDirectSound3DListener interface 
   for that buffer, as shown in the following example with C++ syntax:
*/
 
	ZeroMemory(&dsbdesc, sizeof(DSBUFFERDESC));
	dsbdesc.dwSize = sizeof(DSBUFFERDESC);
	dsbdesc.dwFlags = DSBCAPS_CTRL3D | DSBCAPS_PRIMARYBUFFER;
	if SUCCEEDED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsbPrimary, NULL))
	{
		// Get listener interface
		if FAILED( lpdsbPrimary->QueryInterface(IID_IDirectSound3DListener,
			       (LPVOID *)&lp3dListener))
		{
			lpdsbPrimary->Release();
			return FALSE;
		}
	} 

	lp3dListener->SetPosition( 0.0, 0.0, 0.0, DS3D_DEFERRED );
	lp3dListener->SetOrientation( 1.0, 0.0, 0.0,	// Front Vector
								  0.0, 1.0, 0.0,	// Top Vector
								  DS3D_DEFERRED);
	lp3dListener->SetRolloffFactor( (D3DVALUE) 0.25, DS3D_DEFERRED );
	lp3dListener->SetDistanceFactor( (D3DVALUE).03, DS3D_DEFERRED );
	lp3dListener->CommitDeferredSettings();

	return TRUE;
}


BOOL SoundSet::CreateBuffer( BUFFER_TYPE eBufType )
{
	DSBUFFERDESC    dsbdesc;
	HRESULT			hr;
	BUFFERLIST::iterator it;
	PBUFFERINFO		pBufferInfo;
	lpdsb = NULL;
	switch( eBufType )
	{
	case STATIC:
		memset(&dsbdesc, 0, sizeof(DSBUFFERDESC)); 
		dsbdesc.dwSize = sizeof(DSBUFFERDESC); 
		dsbdesc.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLVOLUME; 
		dsbdesc.dwBufferBytes = mmckinfoData.cksize;  
		dsbdesc.lpwfxFormat = pwfx; 
 
		if FAILED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsb, NULL))
		{
			DISPLAYMSG("Failed to create secondary static buffer");
			return FALSE; 
		}

		// pBufferInfo stores both a sound buffer and its associated 3d buffer
		pBufferInfo = new BUFFERINFO;
		pBufferInfo->lpdsb = lpdsb;
		pBufferInfo->lpdsb3d = NULL;	// Doesn't have associated 3D buffer

		// Add the bufferinfo to the current buffer list
		BufferList.insert( BufferList.end(), pBufferInfo );

		break;

	case STATIC_3D:
		memset(&dsbdesc, 0, sizeof(DSBUFFERDESC)); 
		dsbdesc.dwSize = sizeof(DSBUFFERDESC); 
		dsbdesc.dwFlags = DSBCAPS_CTRL3D | DSBCAPS_STATIC | DSBCAPS_MUTE3DATMAXDISTANCE | 
						  DSBCAPS_CTRLVOLUME;				
		dsbdesc.dwBufferBytes = mmckinfoData.cksize;  
		dsbdesc.lpwfxFormat = pwfx; 
 
		if FAILED(lpds->CreateSoundBuffer(&dsbdesc, &lpdsb, NULL))
		{
			DISPLAYMSG("Failed to create secondary static buffer");
			return FALSE; 
		} 

		hr = lpdsb->QueryInterface(IID_IDirectSound3DBuffer, (LPVOID *)&lpdsb3d); 
		if SUCCEEDED(hr) 
		{ 
			// Set orientation and position of cone
			lpdsb3d->SetPosition( m_vPos.x, m_vPos.y, m_vPos.z, DS3D_IMMEDIATE );
			//lpdsb3d->SetConeOrientation( m_vOrientation.x, m_vOrientation.y, m_vOrientation.z,
			//							 DS3D_IMMEDIATE );
			
		} 

		// pBufferInfo stores both a sound buffer and its associated 3d buffer
		pBufferInfo = new BUFFERINFO;
		pBufferInfo->lpdsb = lpdsb;
		pBufferInfo->lpdsb3d = lpdsb3d;

		// Add the bufferinfo to the current buffer list
		BufferList.insert( BufferList.end(), pBufferInfo );

		break;

	case STREAMING:
		break;

	case STREAMING_3D:
		break;

	default:
		return FALSE;
		break;
	}

	lpdsb->SetVolume( DSBVOLUME_MAX ); 

	return TRUE;
}

BOOL SoundSet::SetListenerPosition(D3DVECTOR vPos)
{
	if( FAILED( lp3dListener->SetPosition( vPos.x, vPos.y, vPos.z, DS3D_IMMEDIATE ) ) ) 
	{
		DISPLAYMSG("Failed to set Listener's Position");
		return FALSE;
	}
	return TRUE;
}

BOOL SoundSet::SetListenerVelocity(D3DVECTOR vVelocity)
{
	if( FAILED( lp3dListener->SetVelocity( vVelocity.x, vVelocity.y, vVelocity.z, 
								   DS3D_IMMEDIATE ) ) )
	{
		DISPLAYMSG("Failed to set Listener's Velocity");
		return FALSE;
	}
	return TRUE;
}

BOOL SoundSet::SetListenerOrientation(D3DVECTOR vFront, D3DVECTOR vTop)
{
	if( FAILED( lp3dListener->SetOrientation( vFront.x, vFront.y, vFront.z,		// Front Vector 
											  vTop.x, vTop.y, vTop.z,			// Top Vector
											  DS3D_IMMEDIATE ) ) ) 
	{
		DISPLAYMSG("Failed to set Listener's Orientation");
		return FALSE;
	}
	return TRUE;
}

// Releases the sound buffers, deletes the bufferinfo and removes it from the list
void SoundSet::ReleaseBufferInfo(PBUFFERINFO pBufInfo)
{
	RELEASE( pBufInfo->lpdsb );		// Release secondary buffer
	RELEASE( pBufInfo->lpdsb3d );	// Release secondary 3D buffer
	delete pBufInfo;				// Deallocate info struct
	BufferList.pop_front( );		// Remove pointer to info struct from list
	
}

BOOL SoundSet::CleanUpBuffers()
{
	PBUFFERINFO pTempBufferInfo;
	DWORD		BufferStatus; 
	// Iterate through buffer list and release all sound buffers

	for( int i = BufferList.size(); i > 0; i-- )
	{
		pTempBufferInfo = BufferList.front();				// Get first info struct
		pTempBufferInfo->lpdsb->GetStatus( &BufferStatus ); // If buffer isn't playing...						
		BufferStatus = BufferStatus & DSBSTATUS_PLAYING;
		if( BufferStatus != DSBSTATUS_PLAYING  )			// ...then remove it.
		{
			ReleaseBufferInfo( pTempBufferInfo );
		}
	}

	return TRUE;
}

BOOL SoundSet::Play(SOUND eSound, D3DVECTOR vPos, D3DVECTOR dOrient, 
					LPLPDIRECTSOUNDBUFFER lplpdsb, LPLPDIRECTSOUND3DBUFFER lplpdsb3d)
{
	Play( eSound, vPos, dOrient );
	*lplpdsb = lpdsb;
	*lplpdsb3d = lpdsb3d;
	
	lpdsb->AddRef();
	lpdsb3d->AddRef();

	return TRUE;
}

BOOL SoundSet::Play(SOUND eSound)
{
	return Play( eSound, m_vPos, m_vOrientation );
	
}

// This function is used for non-3d sounds which wish to have control over the secondary buffer
BOOL SoundSet::Play(SOUND eSound, D3DVECTOR vPos, D3DVECTOR dOrient, LPLPDIRECTSOUNDBUFFER lplpdsb)
{
	
	Play( eSound, vPos, dOrient );
	*lplpdsb = lpdsb;
	
	lpdsb->AddRef();

	return TRUE;
}

BOOL SoundSet::StopBuffer( LPDIRECTSOUNDBUFFER lpdsb )
{
	if( FAILED( lpdsb->Stop() ) )
		return FALSE;
	return TRUE;
}


void SoundSet::ModifyListenerRolloff(D3DVALUE nRolloff)
{
	D3DVALUE newRolloffFactor;

	lp3dListener->GetRolloffFactor( &newRolloffFactor );
	newRolloffFactor *= nRolloff;
	if( newRolloffFactor < 0.1 )
		newRolloffFactor = (D3DVALUE) 0.1;
	else if( newRolloffFactor > 10.0 )
		newRolloffFactor = (D3DVALUE) 10.0;
	lp3dListener->SetRolloffFactor( newRolloffFactor, DS3D_IMMEDIATE );
}
